from contextlib import contextmanager
import logging
import sys

# deprecated
@contextmanager
def log_step(name: str, level=logging.INFO, raise_exception: bool = True):
    try:
        logging.log(level=level, msg=f"Step({name}): Executing...")
        yield
        logging.log(level=level, msg=f"Step({name}): Executed.")
    except Exception as ex:
        logging.error(f"Step({name}): Failed! Reason: {ex}")

        if raise_exception:
            raise

# deprecated
import inspect
from functools import wraps
def logstep(message: str, level=logging.INFO, show_start: bool = True, show_finish: bool = True):
    """
    Decorator to log the start and finish of function execution.

    Args:
        message (str): Custom message to include in logs.
        level (int): Logging level.
        show_start (bool): Whether to log the start of the function.
        show_finish (bool): Whether to log the finish of the function.
    """
    def decorator(func):
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def awrapper(*args, **kwargs_wrapper):
                if show_start:
                    logging.log(level, f"{message} - Starting {func.__name__} with args: {args}, kwargs: {kwargs_wrapper}")
                
                result = await func(*args, **kwargs_wrapper)
                
                if show_finish:
                    logging.log(level, f"{message} - Finished {func.__name__} with result: {result}")
                return result
            return awrapper
        else:
            @wraps(func)
            def wrapper(*args, **kwargs_wrapper):
                if show_start:
                    logging.log(level, f"{message} - Starting {func.__name__} with args: {args}, kwargs: {kwargs_wrapper}")
                
                result = func(*args, **kwargs_wrapper)
                
                if show_finish:
                    logging.log(level, f"{message} - Finished {func.__name__} with result: {result}")
                return result
            return wrapper
    return decorator


def load_test_logger(level = logging.INFO):
  test_logger = logging.getLogger()
  test_logger.setLevel(level=level)
  console_handler = logging.StreamHandler(stream=sys.stdout)
  console_handler.formatter = logging.Formatter('::> %(name)s - %(levelname)s - %(message)s')
  test_logger.handlers.clear()
  test_logger.addHandler(console_handler)


from . import utils as Utils
from .settings.Settings import Settings
from . import errors as Errors
from . import Logs
from . import Decorators

__all__ = [
    'Utils',
    'Settings',
    'Errors',
    'Logs',
    'Decorators'

]



from contextlib import contextmanager
import logging
import inspect
from functools import wraps
from typing import Optional


@contextmanager
def log_step(name: str, level=logging.INFO, raise_exception: bool = True):
    try:
        logging.log(level=level, msg=f"Step({name}): Executing...")
        yield
        logging.log(level=level, msg=f"Step({name}): Executed.")
    except Exception as ex:
        logging.error(f"Step({name}): Failed! Reason: {ex}")

        if raise_exception:
            raise

def logstep(message: str, level=logging.INFO, show_start: bool = True, show_finish: bool = True):
    """
    Decorator to log the start and finish of function execution.

    Args:
        message (str): Custom message to include in logs.
        level (int): Logging level.
        show_start (bool): Whether to log the start of the function.
        show_finish (bool): Whether to log the finish of the function.
    """
    def decorator(func):
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def awrapper(*args, **kwargs_wrapper):
                if show_start:
                    logging.log(level, f"{message} - Starting {func.__name__} with args: {args}, kwargs: {kwargs_wrapper}")
                
                result = await func(*args, **kwargs_wrapper)
                
                if show_finish:
                    logging.log(level, f"{message} - Finished {func.__name__} with result: {result}")
                return result
            return awrapper
        else:
            @wraps(func)
            def wrapper(*args, **kwargs_wrapper):
                if show_start:
                    logging.log(level, f"{message} - Starting {func.__name__} with args: {args}, kwargs: {kwargs_wrapper}")
                
                result = func(*args, **kwargs_wrapper)
                
                if show_finish:
                    logging.log(level, f"{message} - Finished {func.__name__} with result: {result}")
                return result
            return wrapper
    return decorator




from time import perf_counter
from typing import Optional, Callable, TypeVar, Any

F = TypeVar('F', bound=Callable[..., Any])

def timed(_func: Optional[F] = None, *, message: Optional[str] = None, level: int = logging.INFO) -> Callable[..., F]:
    """
    A decorator that logs the time taken by a function to execute.
    Can be used with or without arguments.

    Usage:
        @timed
        def func(...):
            ...

        @timed(message="Custom message", level=logging.DEBUG)
        def func(...):
            ...
    """
    def decorator(func: F) -> F:
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def async_wrapper(*args, **kwargs) -> Any:
                start = perf_counter()
                try:
                    result = await func(*args, **kwargs)
                    return result
                finally:
                    elapsed = perf_counter() - start
                    log_message = f"{func.__name__} took {elapsed:.6f} seconds to finish"
                    if message:
                        log_message = f"{message} - {log_message}"
                    logging.log(level, f"Timed: {log_message}")
            return async_wrapper  # type: ignore
        else:
            @wraps(func)
            def sync_wrapper(*args, **kwargs) -> Any:
                start = perf_counter()
                try:
                    result = func(*args, **kwargs)
                    return result
                finally:
                    elapsed = perf_counter() - start
                    log_message = f"{func.__name__} took {elapsed:.6f} seconds to finish"
                    if message:
                        log_message = f"{message} - {log_message}"
                    logging.log(level, f"Timed: {log_message}")
            return sync_wrapper  # type: ignore

    if _func is None:
        # Decorator is called with arguments
        return decorator
    else:
        # Decorator is used without arguments
        return decorator(_func)



import yaml
import logging
from typing import List, Literal
from ..errors import YamlParseError
import logging
logging.basicConfig(format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")
logger = logging.getLogger(__name__)
import re 
import os

class Settings(dict):
    """
    Represents app settings yaml file
    """

    def __init__(self, section: Literal['llm', 'embeddings', 'vectordb'], path='./app-settings.yaml', **kwargs):
        """
        Reads given yaml file
        - Use kwargs to override values
        """
        self.path = path
        self.section = section

        try:
            # read yaml file
            yaml_content = self._read_yaml(path)

            self._replace_with_environment_variables(yaml_content)

            # add values
            yaml_section = yaml_content[section]
            for key, value in yaml_section.items():
                self[key] = value

            # override values
            self.update({
                key: value
                for key, value in kwargs.items()
            })

        except Exception as ex:
            logger.error(
                f"Failed to read '{section}' section from '{self.path}'. Original: {ex}")
            raise

    def _read_yaml(self, path):
        try:
            with open(path, 'r') as file:
                yaml_content = yaml.safe_load(file)
                return yaml_content
        except FileNotFoundError:
            raise
        except Exception as ex:
            raise YamlParseError(
                f"File '{self.path}' could not be parsed correcty to yaml! Original: {ex}")

    def _set_nested_value(self, yaml_data:dict, keys:List[str], value):
        for key in keys[:-1]:
            yaml_data = yaml_data.setdefault(key, {})
        if value is not None and value:
            yaml_data[keys[-1]] = value

    def _replace_with_environment_variables(self, yaml_content):
        # Pattern to match environment variables with double underscores
        pattern = re.compile(r'(?i)^[a-z][a-z0-9_]*(?:__[a-z][a-z0-9_]*)+$')

        # Process each environment variable
        for key, value in os.environ.items():
            if pattern.match(key):
                # Split the key into parts
                keys = key.lower().split('__')
                # Update the YAML configuration
                self._set_nested_value(yaml_content, keys, value)